import os
from copy import copy

import torch.nn as nn

from abc import abstractmethod

from transformers import (
    XLNetModel,
    AlbertModel,
    BertModel,
    RobertaModel,
    PreTrainedModel,
)
from functools import wraps

from NewsSentiment.download import Download


class FXBaseModel(PreTrainedModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def get_language_models():
        return

    @staticmethod
    @abstractmethod
    def get_input_field_ids():
        return

    def invoke_language_model(self, lm, input_ids, token_type_ids=None):
        type_lm = type(lm)
        if type_lm == XLNetModel:
            last_hidden_state, mems, all_hidden_states = lm(
                input_ids=input_ids, token_type_ids=token_type_ids,
            )
        elif type_lm in [AlbertModel, BertModel, RobertaModel]:
            if token_type_ids is None:
                last_hidden_state, pooler_output, hidden_states = lm(
                    input_ids=input_ids,
                )
            else:
                last_hidden_state, pooler_output, hidden_states = lm(
                    input_ids=input_ids, token_type_ids=token_type_ids
                )
        else:
            raise NotImplementedError

        return last_hidden_state


def provide_pretrained(version, pretrained_url):
    """
    Usage:

        @provide_pretrained("v1.0.0", "https://example.com/link/to/state_dict")
        class Example(nn.Module):
            pass
    """

    def decorator(model_class):
        # The actual decorator to use before the class
        wraps(model_class)

        wrapper = __get_pretrained_wrapper_class(model_class)
        wrapper._provide_pretrained_versions[version] = pretrained_url

        return wrapper

    return decorator


def default_pretrained(version):
    """
    Set the version which should be used as the default version and will be used when running with --pretrained.

    Usage:

        @default_pretrained("v1.0.0")
        @provide_pretrained("v1.0.0", "https://example.com/link/to/state_dict")
        class Example(nn.Module):
            pass
    """

    def decorator(model_class):
        # The actual decorator to use before the class
        wraps(model_class)

        wrapper = __get_pretrained_wrapper_class(model_class)
        wrapper._provide_pretrained_default = version

        return wrapper

    return decorator


def model_includes_pretrained(model):
    """
    Checks if a model-class includes the methods to load pretrained models.

    Arguments:
        model Model-class to check.

    Returns:
        True if it includes the functionality.
    """
    return hasattr(model, "has_pretrained_state_dict") and hasattr(
        model, "get_pretrained_state_dict"
    )


__pretrained_wrapper_classes = set()


def __get_pretrained_wrapper_class(base_class):
    if base_class in __pretrained_wrapper_classes:
        return base_class

    class PretrainedWrapper(base_class):
        _provide_pretrained_default = None
        _provide_pretrained_versions = {}

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @classmethod
        def has_pretrained_state_dict(cls, version=None):
            version = version or cls._provide_pretrained_default
            return version in cls._provide_pretrained_versions

        @classmethod
        def get_pretrained_state_dict(
            cls, version=None, download_if_not_exists=True, **kwargs
        ):
            path = Download.model_path(cls, version)
            if os.path.isfile(path):
                if download_if_not_exists:
                    Download.download(cls, version, False)
                else:
                    raise FileNotFoundError("State dict not found")
            return cls.load_state_dict(path, **kwargs)

        @classmethod
        def get_pretrained_versions(cls):
            return copy(cls._provide_pretrained_versions)

        @classmethod
        def get_pretrained_source(cls, version=None):
            return cls._provide_pretrained_versions[
                version or cls._provide_pretrained_default
            ]

        @classmethod
        def get_pretrained_default_version(cls):
            return cls._provide_pretrained_default

    __pretrained_wrapper_classes.add(PretrainedWrapper)
    return PretrainedWrapper



"""
This file converts the dataset files (3 splits) into the format we'll use on
Huggingface Hub, i.e., where examples with k targets are expanded to k examples, each
having 1 target.
"""

import pathlib
import jsonlines
from loguru import logger


def convert_polarity(polarity):
    if polarity == 2.0:
        return -1
    elif polarity == 4.0:
        return 0
    elif polarity == 6.0:
        return 1
    else:
        raise ValueError


def convert_target(obj, target):
    converted_obj = {
        "mention": target["mention"],
        "polarity": convert_polarity(target["polarity"]),
        "from": target["from"],
        "to": target["to"],
        "sentence": obj["sentence_normalized"],
        "id": target["Input.gid"],
    }

    return converted_obj


def convert_obj(obj):
    targets = obj["targets"]
    converted_objs = []

    for target in targets:
        converted_objs.append(convert_target(obj, target))

    return converted_objs


def convert(path):
    files = [p for p in pathlib.Path(path).iterdir() if p.is_file()]

    for file in files:
        converted_lines = []
        counter = 0
        with jsonlines.open(file) as reader:
            for obj in reader:
                converted_lines.extend(convert_obj(obj))
                counter += 1
        logger.info(
            "converted {} lines to {} examples in {}",
            counter,
            len(converted_lines),
            file,
        )

        with jsonlines.open(str(file) + "converted", mode="w") as writer:
            writer.write_all(converted_lines)


if __name__ == "__main__":
    convert("experiments/default/datasets/newsmtsc-mt")
    convert("experiments/default/datasets/newsmtsc-rw")
